from .converters import (
    integer,
    number,
    string,
    boolean,
)

from .validators import (
    required,
    length,
    regex,
)

from .schema import (
    array,
    dictionary,
    process,
)

from .exc import (
    SchematecError,
    ConvertationError,
    ValidationError,
    SchemaError,
)

__all__ = [
    'integer',
    'number',
    'string',
    'boolean',
    'required',
    'length',
    'regex',
    'array',
    'dictionary',
    'process',
    'SchematecError',
    'ConvertationError',
    'ValidationError',
    'SchemaError',
]

__version__ = '0.4.1'



from __future__ import absolute_import

import schematec.abc as abc
import schematec.converters
import schematec.exc as exc


class Dictionary(abc.Schema):
    def __init__(self, **descriptors):
        self.descriptors = descriptors

    def __call__(self, data, strict=False):
        data = schematec.converters.dictionary(data)

        if not self.descriptors:
            return data

        result = {}

        for name, descriptors in self.descriptors.items():
            if isinstance(descriptors, abc.AbstractDescriptor):
                descriptors = [descriptors]
            elif isinstance(descriptors, abc.ComplexDescriptor):
                descriptors = list(descriptors)
            else:
                raise TypeError(descriptors)

            unbound_validators = [v for v in descriptors if
                                  isinstance(v, abc.Validator) and not v.BINDING]
            for validator in unbound_validators:
                validator(name, data)

            try:
                value = data[name]
            except KeyError:
                if strict:
                    raise exc.SchemaError(name)
                else:
                    continue

            schemas = [s for s in descriptors if isinstance(s, abc.Schema)]
            for schema in schemas:
                value = schema(value, strict=strict)

            converters = [c for c in descriptors if isinstance(c, abc.Converter)]
            for converter in converters:
                value = converter(value)

            bound_validators = [v for v in descriptors if isinstance(v, abc.Validator) and v.BINDING]
            for validator in bound_validators:
                if isinstance(value, validator.BINDING):
                    validator(value)
            result[name] = value

        return result

dictionary = Dictionary


class Array(abc.Schema):
    def __init__(self, *descriptors):
        if descriptors:
            descriptors = descriptors[0]
            if isinstance(descriptors, abc.AbstractDescriptor):
                self.descriptors = [descriptors]
            elif isinstance(descriptors, abc.ComplexDescriptor):
                self.descriptors = list(descriptors)
            else:
                raise TypeError(descriptors)
        else:
            self.descriptors = []

    def __call__(self, data, strict=False):
        data = schematec.converters.array(data)

        if not self.descriptors:
            return data

        schemas = [s for s in self.descriptors if isinstance(s, abc.Schema)]
        for schema in schemas:
            data = [schema(d, strict=strict) for d in data]

        converters = [c for c in self.descriptors if isinstance(c, abc.Converter)]
        for converter in converters:
            data = map(converter, data)

        bound_validators = [v for v in self.descriptors if isinstance(v, abc.Validator) and v.BINDING]
        for validator in bound_validators:
            for value in data:
                if isinstance(value, validator.BINDING):
                    validator(value)

        return data

array = Array


def expand(schema):
    if isinstance(schema, (abc.AbstractDescriptor, abc.ComplexDescriptor)):
        return schema
    elif isinstance(schema, dict):
        return Dictionary(**{n: expand(d) for n, d in schema.items()})
    elif isinstance(schema, list):
        if schema:
            return Array(expand(schema[0]))
        else:
            return Array()
    else:
        raise exc.SchemaError(schema)


def process(schema, data, strict=False):
    return expand(schema)(data, strict=strict)



from __future__ import absolute_import


class Descriptor(object):
    pass


class ComplexDescriptor(Descriptor):
    def __init__(self, *descriptors):
        self.descriptors = list(descriptors)

    def __and__(self, descriptor):
        if not isinstance(descriptor, AbstractDescriptor):
            raise TypeError(descriptor)

        self.descriptors.append(descriptor)
        return self

    def __iter__(self):
        for descriptor in self.descriptors:
            yield descriptor


class AbstractDescriptor(Descriptor):
    def __init__(self):
        pass

    def __call__(self, *args, **kw):
        raise NotImplementedError

    def __and__(self, descriptor):
        if not isinstance(descriptor, AbstractDescriptor):
            raise TypeError(descriptor)

        return ComplexDescriptor(self, descriptor)


class Schema(AbstractDescriptor):
    pass


class Converter(AbstractDescriptor):
    pass


class Validator(AbstractDescriptor):
    BINDING = None



from __future__ import absolute_import

import collections

import schematec.exc as exc
import schematec.abc as abc


class Integer(abc.Converter):
    def __call__(self, value):
        if value is None:
            raise exc.ConvertationError(value)

        if isinstance(value, bool):
            return int(value)

        if isinstance(value, (int, long)):
            return int(value)

        if isinstance(value, basestring):
            try:
                return int(value)
            except ValueError:
                raise exc.ConvertationError(value)

        raise exc.ConvertationError(value)

integer = Integer()


class Number(abc.Converter):
    def __call__(self, value):
        if value is None:
            raise exc.ConvertationError(value)

        if isinstance(value, bool):
            return float(value)

        if isinstance(value, (float, int, long)):
            return float(value)

        if isinstance(value, basestring):
            try:
                return float(value)
            except ValueError:
                raise exc.ConvertationError(value)

        raise exc.ConvertationError(value)

number = Number()


class String(abc.Converter):
    def __call__(self, value):
        if value is None:
            raise exc.ConvertationError(value)

        if isinstance(value, unicode):
            return value

        if isinstance(value, bool):
            raise exc.ConvertationError(value)

        if isinstance(value, (int, long)):
            return unicode(value)

        if isinstance(value, str):
            try:
                return unicode(value)
            except UnicodeDecodeError:
                raise exc.ConvertationError(value)

        raise exc.ConvertationError(value)

string = String()


class Boolean(abc.Converter):
    def __call__(self, value):
        if value is None:
            raise exc.ConvertationError(value)

        if isinstance(value, bool):
            return value

        if isinstance(value, (int, long)) and value in (0, 1):
            return bool(value)

        if isinstance(value, basestring) and value in (u'0', u'1'):
            return bool(int(value))

        raise exc.ConvertationError(value)

boolean = Boolean()


class Array(abc.Converter):
    def __call__(self, value):
        if isinstance(value, collections.Iterable) and not isinstance(value, collections.Mapping):
            return list(value)

        raise exc.ConvertationError(value)

array = Array()


class Dictionary(abc.Converter):
    def __call__(self, value):
        if isinstance(value, collections.Mapping):
            return dict(value)

        raise exc.ConvertationError(value)

dictionary = Dictionary()



from __future__ import absolute_import

import collections

import schematec.exc as exc
import schematec.abc as abc


class Required(abc.Validator):
    def __call__(self, name, data):
        if name not in data:
            raise exc.ValidationError(name)

required = Required()


class Length(abc.Validator):
    BINDING = (str, collections.Sized)

    def __init__(self, max_length):
        self.max_length = max_length

    def __call__(self, value):
        if len(value) > self.max_length:
            raise exc.ValidationError(value)

length = Length


class Regex(abc.Validator):
    BINDING = (str, )

    def __init__(self, regex):
        self.regex = regex

    def __call__(self, value):
        if not self.regex.match(value):
            raise exc.ValidationError(value)

regex = Regex



import re
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = ''

with open('schematec/__init__.py', 'r') as fd:
    regex = re.compile(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]')
    for line in fd:
        m = regex.match(line)
        if m:
            version = m.group(1)
            break

setup(
    name='schematec',
    packages=['schematec'],
    package_data={'': ['LICENSE']},
    version=version,
    description='Set of tools that makes input data validation easier',
    author='Andrey Gubarev',
    author_email='mylokin@me.com',
    url='https://github.com/mylokin/schematec',
    keywords=['schema'],
    license='MIT',
    classifiers=(
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Topic :: Database',
        'Topic :: Database :: Database Engines/Servers',
    ),
)
